@@ -16,16 +16,30 @@ import os
 from collections import defaultdict
 
 import pytest
-from pytest_terraform import tf, xdist
+from pytest_terraform import hooks, tf, xdist
 
 
 @pytest.hookimpl(trylast=True)
 def pytest_configure(config):
-    tf.LazyTfBin.value = config.getoption("dest_tf_binary") or tf.find_binary("terraform")
     tf.LazyPluginCacheDir.value = cache_dir = config.getoption("dest_tf_plugin")
     if not os.path.exists(cache_dir):
         os.mkdir(cache_dir)
-    tf.LazyReplay.value = config.getoption("dest_tf_replay")
+
+    tf.LazyModuleDir.value = config.getoption("dest_tf_mod_dir") or config.getini(
+        "terraform-mod-dir"
+    )
+
+    if tf.LazyReplay.value is None:
+        tf.LazyReplay.value = config.getoption("dest_tf_replay")
+
+    tf.LazyTfBin.value = config.getoption("dest_tf_binary") or tf.find_binary("terraform")
+    if tf.LazyTfBin.value is None and not tf.LazyReplay.value:
+        raise ValueError(
+            "pytest-terraform requires terraform binary on PATH or "
+            "specified with --tf-binary"
+        )
+
+    tf.PytestConfig.value = config
 
     if config.pluginmanager.hasplugin("xdist"):
         config.pluginmanager.register(xdist.XDistTerraform(config))
@@ -35,6 +49,11 @@ def pytest_configure(config):
         d["function"] = tf.TerraformFixture
 
 
+def pytest_addhooks(pluginmanager):
+    """ Register pytest_terraform hooks """
+    pluginmanager.add_hookspecs(hooks)
+
+
 def pytest_addoption(parser):
     group = parser.getgroup("terraform")
     group.addoption(
