@@ -16,12 +16,16 @@ import json
 import os
 import subprocess
 import sys
-from collections import defaultdict
+from collections import UserString, defaultdict
+from typing import Any, Dict, Optional, Tuple, Union
 
 import jmespath
 import pytest
 from py.path import local
 
+from .exceptions import InvalidState, ModuleNotFound, TerraformCommandFailed
+from .options import teardown as td
+
 
 def find_binary(bin_name):
     parts = os.environ["PATH"].split(":")
@@ -31,10 +35,6 @@ def find_binary(bin_name):
             return candidate
 
 
-class ModuleNotFound(ValueError):
-    """module not found"""
-
-
 class TerraformRunner(object):
 
     command_templates = {
@@ -80,7 +80,7 @@ class TerraformRunner(object):
         elif plan:
             apply_args = self._get_cmd_args("apply", plan="")
         self._run_cmd(apply_args)
-        return TerraformState.load(self.state_path)
+        return TerraformState.from_file(self.state_path)
 
     def plan(self, output=""):
         output = output and "-out=%s" % output or ""
@@ -113,6 +113,39 @@ class TerraformRunner(object):
         run_cmd(args, cwd=cwd, stderr=subprocess.STDOUT, env=env)
 
 
+class TerraformStateJson(UserString):
+    @classmethod
+    def from_dict(cls, state: Dict[str, Any]):
+        """create TerraformStateJson from dictionary"""
+        s = cls("")
+        s.update_dict(state)
+        return s
+
+    def update(self, state: str):
+        """update TerraformStateJson object with new data"""
+        if not isinstance(state, str):
+            raise ValueError(f"{state} is not a string")
+
+        self.data = str(state)
+
+    def update_dict(self, state: Dict[str, Any]):
+        """update TerraformStateJson from a dict"""
+        self.update(json.dumps(state, indent=4))
+
+    @property
+    def dict(self):
+        """return the TerraformStateJson as a dict"""
+        return json.loads(self.data)
+
+    @dict.setter
+    def dict(self, data: Dict[str, Any]):
+        """update TerraformStateJson from a dict"""
+        try:
+            self.update_dict(data)
+        except (ValueError, TypeError):
+            raise ValueError("Not a serializable object")
+
+
 class TerraformState(object):
     """Abstraction over a terrafrom state file with helpers.
 
@@ -164,41 +197,92 @@ class TerraformState(object):
         return default
 
     @classmethod
-    def load(cls, state_file):
-        resources = {}
-        outputs = {}
-        with open(state_file) as fh:
-            data = json.load(fh)
-            if "pytest-terraform" in data:
-                return cls(data["resources"], data["outputs"])
-            for r in data.get("resources", ()):
-                rmap = resources.setdefault(r["type"], {})
-                rmap[r["name"]] = dict(r["instances"][0]["attributes"])
+    def from_file(cls, path: str):
+        """create TerraformState from a file
 
-            outputs = data.get("outputs", {})
-            for m in data.get("modules", ()):
-                for k, r in m.get("resources", {}).items():
-                    if k.startswith("data"):
-                        continue
-                    module, rname = k.split(".", 1)
-                    rmap = resources.setdefault(module, {})
-                    rattrs = {"id": r["primary"]["id"]}
-                    for kattr, vattr in r["primary"]["attributes"].items():
-                        if "name" in kattr and vattr != rattrs["id"]:
-                            rattrs[kattr] = vattr
-                    rmap[rname] = rattrs
+        File can either be a Terraform Plan state, or a recorded
+        pytest-terraform state
+        """
+        if not os.path.isfile(path):
+            raise InvalidState("{} could not be located".format(path))
+
+        with open(path) as fh:
+            state = fh.read()
+
+        return cls.from_string(state)
+
+    @classmethod
+    def from_string(cls, state: Union[TerraformStateJson, str]):
+        """create TerraformState from string
+
+        State string can be a bytestring or a TerraformStateJson
+        string object
+        """
+        resources, outputs = cls.parse_state(state)
         return cls(resources, outputs)
 
-    def save(self, state_path):
+    def update(self, state: Union[TerraformStateJson, str]):
+        """update TerraformState values"""
+        resources, outputs = self.parse_state(state)
+        self.resources = resources
+        self.outputs = outputs
+
+    @staticmethod
+    def parse_state(
+        state: Union[TerraformStateJson, str]
+    ) -> Tuple[Dict[str, any], Dict[str, Any]]:
+        """extract resources and outputs from state
+
+        where state is one of the following:
+        * Terraform state output as a string
+        * Recorded pytest-terraform state
+        * TerraformStateJson object
+        """
+        if isinstance(state, TerraformStateJson):
+            data = state.dict
+        else:
+            data = json.loads(state)
+
+        if "pytest-terraform" in data:
+            return (data["resources"], data["outputs"])
+
+        resources = {}
+        outputs = {}
+
+        for r in data.get("resources", ()):
+            rmap = resources.setdefault(r["type"], {})
+            rmap[r["name"]] = dict(r["instances"][0]["attributes"])
+
+        outputs = data.get("outputs", {})
+        for m in data.get("modules", ()):
+            for k, r in m.get("resources", {}).items():
+                if k.startswith("data"):
+                    continue
+                module, rname = k.split(".", 1)
+                rmap = resources.setdefault(module, {})
+                rattrs = {"id": r["primary"]["id"]}
+                for kattr, vattr in r["primary"]["attributes"].items():
+                    if "name" in kattr and vattr != rattrs["id"]:
+                        rattrs[kattr] = vattr
+                rmap[rname] = rattrs
+
+        return (resources, outputs)
+
+    def save(self, state_path: Optional[str] = None) -> Optional[TerraformStateJson]:
+        """export state as a string or to a file"""
+        state = {
+            "pytest-terraform": 1,
+            "outputs": self.outputs,
+            "resources": self.resources,
+        }
+
+        output = TerraformStateJson.from_dict(state)
+
+        if not state_path:
+            return output
+
         with open(state_path, "w") as fh:
-            json.dump(
-                {
-                    "pytest-terraform": 1,
-                    "outputs": self.outputs,
-                    "resources": self.resources,
-                },
-                fh,
-            )
+            fh.write(str(output))
 
 
 class TerraformTestApi(TerraformState):
@@ -211,7 +295,7 @@ class PlaceHolderValue(object):
     many of our instantiations are at module import time, to support
     runtime configuration from cli/ini options we utilize a lazy
     loaded value set which is configured for final values via hooks
-    (early, post conf, pre collection).)
+    (early, post conf, pre collection).
     """
 
     def __init__(self, name):
@@ -228,19 +312,29 @@ LazyReplay = PlaceHolderValue("tf_replay")
 LazyModuleDir = PlaceHolderValue("module_dir")
 LazyPluginCacheDir = PlaceHolderValue("plugin_cache")
 LazyTfBin = PlaceHolderValue("tf_bin_path")
+PytestConfig = PlaceHolderValue("pytestconfig")
 
 
 class TerraformFixture(object):
-
-    _AutoTearDown = True
-
-    def __init__(self, tf_bin, plugin_cache, scope, tf_root_module, test_dir, replay):
+    def __init__(
+        self,
+        tf_bin,
+        plugin_cache,
+        scope,
+        tf_root_module,
+        test_dir,
+        replay,
+        teardown,
+        pytest_config,
+    ):
         self.tf_bin = tf_bin
         self.tf_root_module = tf_root_module
         self.test_dir = test_dir
         self.scope = scope
         self.replay = replay
         self.runner = None
+        self.teardown_config = td.resolve(teardown)
+        self.config = pytest_config
 
     @property
     def name(self):
@@ -279,7 +373,9 @@ class TerraformFixture(object):
                 raise ValueError(
                     "Replay resources don't exist for %s" % self.tf_root_module
                 )
-            return TerraformTestApi.load(os.path.join(module_dir, "tf_resources.json"))
+            return TerraformTestApi.from_file(
+                os.path.join(module_dir, "tf_resources.json")
+            )
         work_dir = tmpdir_factory.mktemp(self.tf_root_module, numbered=True).join("work")
         self.runner = self.get_runner(module_dir, work_dir)
         return self.create(request, module_dir)
@@ -287,10 +383,13 @@ class TerraformFixture(object):
     def create(self, request, module_dir):
         print("tf create %s" % self.tf_root_module, file=sys.stderr)
         self.runner.init()
-        if self._AutoTearDown:
+        if self.teardown_config != td.OFF:
             request.addfinalizer(self.tear_down)
         try:
             test_api = self.runner.apply()
+            tfstatejson = test_api.save()
+            self.config.hook.pytest_terraform_modify_state(tfstate=tfstatejson)
+            test_api.update(tfstatejson)
             test_api.save(module_dir.join("tf_resources.json"))
             return test_api
         except Exception:
@@ -299,7 +398,12 @@ class TerraformFixture(object):
     def tear_down(self):
         # config behavor on runner
         print("tf teardown %s" % self.tf_root_module, file=sys.stderr)
-        self.runner.destroy()
+        try:
+            self.runner.destroy()
+        except subprocess.CalledProcessError as e:
+            if self.teardown_config == td.IGNORE:
+                return
+            raise TerraformCommandFailed from e
 
 
 class FixtureDecoratorFactory(object):
@@ -320,7 +424,14 @@ class FixtureDecoratorFactory(object):
                 return f
         raise KeyError(name)
 
-    def __call__(self, terraform_dir, scope="function", replay=None, name=None):
+    def __call__(
+        self,
+        terraform_dir,
+        scope="function",
+        replay=None,
+        name=None,
+        teardown=td.DEFAULT,
+    ):
         # We have to hook into where fixture discovery will find
         # our fixtures, the easiest option is to store on the module that
         # originated the call, all test modules get scanned for
@@ -344,7 +455,14 @@ class FixtureDecoratorFactory(object):
             return self.nonce_decorator
         tclass = self.scope_class_map[scope]
         tfix = tclass(
-            LazyTfBin, LazyPluginCacheDir, scope, terraform_dir, test_dir, replay,
+            LazyTfBin,
+            LazyPluginCacheDir,
+            scope,
+            terraform_dir,
+            test_dir,
+            replay,
+            teardown,
+            PytestConfig.resolve(),
         )
         self._fixtures.append(tfix)
         marker = pytest.fixture(scope=scope, name=terraform_dir)
